// Written by ChatGPT 3.5 with help of Kaze (sanmayce@sanmayce.com), hee-hee.
// How to compile: $ gcc -O3 -o bandwidth_XMM.elf -fopenmp bandwidth_XMM.c -msse4.2
// How to run: $ /bin/time -v ./bandwidth_XMM.elf 
// With above run you can be sure the memory footprint to be 1GB*4, this is done for machinettes with 8GB.

#include <stdio.h>
#include <stdlib.h>
#include <x86intrin.h> // Include Intel Intrinsics
#include <omp.h>
#include <time.h> // Include for time()

//#define ARRAY_SIZE (1 << 30) // 1GB array size
size_t ARRAY_SIZE = 1*1024*1024*1024ULL; // 1GB * 4 in order to fit into 8 GB
#define NUM_RUNS 5

// Function to initialize the array with random values
void initializeArray(float *array, size_t size) {
    printf("Initializing block %d MB ...\n", (long long)(double)(size * sizeof(float))>>20);
    #pragma omp parallel for
    for (size_t i = 0; i < size; ++i) {
        array[i] = (float)rand() / RAND_MAX;
    }
}

// Function to benchmark RAM read bandwidth
double benchmarkBandwidth(float *array, size_t size) {
    // Initialize vectors
    __m128 vec;
    float sum_arr[4] = {0.0f, 0.0f, 0.0f, 0.0f};

    // Measure the start time
    double startMP = omp_get_wtime();
    time_t start, end;
    clock_t startCLCK, endCLCK;
    time(&start);
    startCLCK = clock();

    // Vectorized memory read
    #pragma omp parallel for
    for (size_t i = 0; i < size; i += 4) {
        vec = _mm_loadu_ps(&array[i]);
        float temp[4];
        _mm_storeu_ps(temp, vec);
        #pragma omp simd
        for (int j = 0; j < 4; j++) {
            sum_arr[j] += temp[j];
        }
    }

    // Perform horizontal sum
    float total_sum = 0.0f;
    for (int j = 0; j < 4; j++) {
        total_sum += sum_arr[j];
    }

    // Measure the end time
    double endMP = omp_get_wtime();
    time(&end);
    endCLCK = clock();

    // Calculate bandwidth
    double timew = (endMP - startMP);
    double timeC = (end - start);
    double timeCLCK = ((double)(endCLCK - startCLCK)) / CLOCKS_PER_SEC;
    double bytes = (double)(size * sizeof(float));
    double bandwidth = bytes / timew;

    printf("omp_get_wtime(): %.2f seconds\n", timew);
    //printf("C standard library time(): %.4f seconds\n", timeC);
    //printf("C standard library clock(): %.2f seconds\n", timeCLCK);

    return bandwidth;
}

int main() {
    int MAXthreads;
    int MINthreads;
    // Allocate memory for the array
    float *array = (float *)malloc(ARRAY_SIZE * sizeof(float));
    if (array == NULL) {
        printf("Memory allocation failed.\n");
        return 1;
    }

    // Initialize the array with random values
    initializeArray(array, ARRAY_SIZE);
    MAXthreads=omp_get_max_threads();
    MINthreads=1;//MAXthreads;
	printf("Cycling through %d to %d threads...\n", MINthreads, MAXthreads);

    // Benchmark the RAM read bandwidth
    for (int num_threads = MINthreads; num_threads <= MAXthreads; ++num_threads) {
        omp_set_num_threads(num_threads);
        double max_bandwidth = 0.0;

        // Warm-up cache
/*
        for (int i = 0; i < 100; ++i) {
            double bandwidth = benchmarkBandwidth(array, ARRAY_SIZE);
            if (bandwidth > max_bandwidth) {
                max_bandwidth = bandwidth;
            }
        }
*/

        // Benchmark the RAM read bandwidth
        for (int run = 0; run < NUM_RUNS; ++run) {
            double bandwidth = benchmarkBandwidth(array, ARRAY_SIZE);
            if (bandwidth > max_bandwidth) {
                max_bandwidth = bandwidth;
            }
        }

        // Print the maximum observed bandwidth for the current num_threads scenario
        printf("Threads: %d, Maximum observed bandwidth: %.2f GB/s\n", num_threads, max_bandwidth / (1e9));
    }

    // Free allocated memory
    free(array);

    return 0;
}


