// Reports FNPS (Factorized-Numbers-Per-Second) calculated for 4096 numbers.
// $ gcc -O3 factorize_FNPS.c -o factorize_FNPS.elf -fopenmp
// $ clang -O3 factorize_FNPS.c -o factorize_FNPS.elf -fopenmp
// Note: Once again, clang is much faster?!

/*
___________               .__   .__     .___.__   __     __          
\_   _____/ __ __   ____  |  |  |__|  __| _/|__|_/  |_ _/  |_  ____  
 |    __)_ |  |  \_/ ___\ |  |  |  | / __ | |  |\   __\\   __\/  _ \ 
 |        \|  |  /\  \___ |  |__|  |/ /_/ | |  | |  |   |  | (  <_> )
/_______  /|____/  \___  >|____/|__|\____ | |__| |__|   |__|  \____/ 
        \/             \/                \/                          

[Fundamental Theorem of Arithmetic Statement]

Every integer greater than 1 is either a prime number, or can be expressed uniquely as a product of prime numbers (the factorization is unique except for the order).

[Why Is the Fundamental Theorem of Arithmetic Important?]

- The theorem forms the foundation for many number theory concepts and allows us to solve a wide range of mathematical problems.
- It allows us to break down any number into its prime factors using Prime Factorization, providing valuable information about its properties and structure.
- By decomposing a number into its prime factors, we can easily determine if it is divisible by another number and identify its factors.
- The theorem guarantees that the prime factorization of a number is unique. This means that two different numbers cannot have the same prime factorization
*/

#define NofN 4096
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <time.h>
#include <omp.h>

void factorize(uint64_t n) {
    uint64_t i;
    char buffer[1024];
    int offset = sprintf(buffer, "%llu: ", (unsigned long long)n);
    for (i = 2; i <= n; i++) {
        while (n % i == 0) {
            offset += sprintf(buffer + offset, "%llu", (unsigned long long)i);
            n /= i;
            if (n > 1)
                offset += sprintf(buffer + offset, "x");
        }
    }
    sprintf(buffer + offset, "\n");

    #pragma omp critical
    {
        printf("%s", buffer);
    }
}

void factorizeOLD(uint64_t n) {
    uint64_t i;
    #pragma omp critical
	{
	    printf("%llu: ", (unsigned long long)n);
		for (i = 2; i <= n; i++) {
		    while (n % i == 0) {
		        printf("%llu", (unsigned long long)i);
		        n /= i;
		        if (n > 1) printf("x");
		    }
		}
		printf("\n");
	}
}

int main() {
    uint64_t i;
    uint64_t max_uint64 = UINT64_MAX;
    uint64_t istart=1999999999;
    uint64_t iend=1999999999+(4*NofN-1);

	//clock_t start_time, end_time;
	//start_time = clock();
	//double cpu_time_used=0;

		double start_time;
	    double end_time;
	    double elapsed_time;
		int num_threads;

    #pragma omp parallel
    {
		// Getting the number of threads
	    num_threads = omp_get_num_threads();
		//printf("Number of threads: %d\n", num_threads);
		omp_set_num_threads(num_threads);
	}
		printf("Benchmarking scalar factorization with OpenMP using %d threads...:\n", num_threads);
    #pragma omp parallel
    {
		start_time = omp_get_wtime();
		#pragma omp for
			//for (i = 1; i <= max_uint64; i++) {
			for (i = istart; i <= iend; i++) {
				factorize(i);
			/*
					if (i > 13,000,000) { 
						end_time = clock();
						cpu_time_used = ((double) (end_time - start_time)) / CLOCKS_PER_SEC;
						printf("Factorized 13,000,000 mumbers, %.2f FNPS.\n", i/cpu_time_used); exit(0);
					}
			*/
				//if (i % 3333 == 0) {
				    //end_time = clock();
				    //cpu_time_used = ((double) (end_time - start_time)) / CLOCKS_PER_SEC;
				    //cpu_time_used = (i-istart)/cpu_time_used;
				//}
			}
    }

    end_time = omp_get_wtime();
    elapsed_time = end_time - start_time;
    printf("Elapsed time: %.2f seconds\n", elapsed_time);
    printf("Factorizations per second: %.2f\n", (iend-istart+1) / elapsed_time);

    return 0;
}

/*
[sanmayce@dzvertcheto _KAZE_maintenance_utilities]$ sh BENCHMARK_4096.sh 
Architecture:            x86_64
  CPU op-mode(s):        32-bit, 64-bit
  Address sizes:         39 bits physical, 48 bits virtual
  Byte Order:            Little Endian
CPU(s):                  8
  On-line CPU(s) list:   0-7
Vendor ID:               GenuineIntel
  Model name:            Intel(R) Core(TM) i7-8565U CPU @ 1.80GHz
    CPU family:          6
    Model:               142
    Thread(s) per core:  2
    Core(s) per socket:  4
    Socket(s):           1
    Stepping:            12
    CPU(s) scaling MHz:  82%
    CPU max MHz:         4600.0000
    CPU min MHz:         400.0000
    BogoMIPS:            3999.93
    Flags:               fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc art arch_perfmon pebs bts rep_good nopl xtopology
                          nonstop_tsc cpuid aperfmperf pni pclmulqdq dtes64 monitor ds_cpl vmx est tm2 ssse3 sdbg fma cx16 xtpr pdcm pcid sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm 3dnowpr
                         efetch cpuid_fault epb ssbd ibrs ibpb stibp ibrs_enhanced tpr_shadow flexpriority ept vpid ept_ad fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid mpx rdseed adx smap clflushopt intel_pt xsaveopt xsavec xget
                         bv1 xsaves dtherm ida arat pln pts hwp hwp_notify hwp_act_window hwp_epp vnmi md_clear flush_l1d arch_capabilities
Virtualization features: 
  Virtualization:        VT-x
Caches (sum of all):     
  L1d:                   128 KiB (4 instances)
  L1i:                   128 KiB (4 instances)
  L2:                    1 MiB (4 instances)
  L3:                    8 MiB (1 instance)
NUMA:                    
  NUMA node(s):          1
  NUMA node0 CPU(s):     0-7
Vulnerabilities:         
  Gather data sampling:  Mitigation; Microcode
  Itlb multihit:         KVM: Mitigation: VMX disabled
  L1tf:                  Not affected
  Mds:                   Not affected
  Meltdown:              Not affected
  Mmio stale data:       Mitigation; Clear CPU buffers; SMT vulnerable
  Retbleed:              Mitigation; Enhanced IBRS
  Spec rstack overflow:  Not affected
  Spec store bypass:     Mitigation; Speculative Store Bypass disabled via prctl
  Spectre v1:            Mitigation; usercopy/swapgs barriers and __user pointer sanitization
  Spectre v2:            Mitigation; Enhanced / Automatic IBRS, IBPB conditional, RSB filling, PBRSB-eIBRS SW sequence
  Srbds:                 Mitigation; Microcode
  Tsx async abort:       Not affected
Benchmarking scalar factorization with OpenMP using 8 threads...:
2000014335: 3x3x3x3x5x11x23x131x149
2000012287: 7x13x179x199x617
2000008191: 3x19x109x487x661
2000008192: 2x2x2x2x2x2x2x2x2x2x7x13x13x13x127
2000014336: 2x2x2x2x2x2x2x2x2x2x71x27509
2000012288: 2x2x2x2x2x2x2x2x2x2x23x84919
2000006143: 21713x92111
...
2000010234: 2x3x333335039
2000010235: 5x400002047
2000010236: 2x2x7x71428937
2000010237: 3x17x257x331x461
2000010238: 2x43x2903x8011
Elapsed time: 2186.81 seconds
Factorizations per second: 7.49

 Performance counter stats for './factorize_FNPS.elf':

     16,551,134.51 msec task-clock:u                     #    7.569 CPUs utilized             
                 0      context-switches:u               #    0.000 /sec                      
                 0      cpu-migrations:u                 #    0.000 /sec                      
               249      page-faults:u                    #    0.015 /sec                      
54,625,440,848,086      cycles:u                         #    3.300 GHz                         (50.00%)
33,850,073,366,914      instructions:u                   #    0.62  insn per cycle              (62.50%)
 7,813,309,701,850      branches:u                       #  472.071 M/sec                       (62.50%)
         4,802,288      branch-misses:u                  #    0.00% of all branches             (62.50%)
       562,959,894      L1-dcache-loads:u                #   34.013 K/sec                       (62.50%)
        21,497,779      L1-dcache-load-misses:u          #    3.82% of all L1-dcache accesses   (62.50%)
         5,634,773      LLC-loads:u                      #  340.446 /sec                        (50.00%)
           832,103      LLC-load-misses:u                #   14.77% of all L1-icache accesses   (50.00%)

    2186.810700687 seconds time elapsed

   16426.297672000 seconds user
       6.522060000 seconds sys

[sanmayce@dzvertcheto _KAZE_maintenance_utilities]$ 
*/
